#include <stdio.h>
#include <stdlib.h>
#include <wiringPi.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <getopt.h>
#include <string.h>

// #define LED 0 matches with ASUS_GPIO 164! This can be checked with command 'sudo gpio readall'.
#define LED    3
/*
ASUS_TINKERBOARD:/ # gpio readall
 +-----+-----+---------+------+---+--Tinker--+---+------+---------+-----+-----+
 | CPU | wPi |   Name  | Mode | V | Physical | V | Mode | Name    | wPi | CPU |
 +-----+-----+---------+------+---+----++----+---+------+---------+-----+-----+
 |     |     |    3.3v |      |   |  1 || 2  |   |      | 5v      |     |     |
 | 252 |   8 |   SDA.1 |      | 1 |  3 || 4  |   |      | 5V      |     |     |
 | 253 |   9 |   SCL.1 |      | 1 |  5 || 6  |   |      | 0v      |     |     |
 |  17 |   7 | GPIO0C1 |  OUT | 0 |  7 || 8  | 1 | SERL | TxD1    | 15  | 161 |
 |     |     |      0v |      |   |  9 || 10 | 1 | SERL | RxD1    | 16  | 160 |
 | 164 |   0 | GPIO5B4 |  OUT | 1 | 11 || 12 | 0 |      | GPIO6A0 | 1   | 184 |
 | 166 |   2 | GPIO5B6 |   IN | 0 | 13 || 14 |   |      | 0v      |     |     |
 | 167 |   3 | GPIO5B7 |   IN | 1 | 15 || 16 | 1 | IN   | GPIO5B2 | 4   | 162 |
 |     |     |    3.3v |      |   | 17 || 18 | 1 | IN   | GPIO5B3 | 5   | 163 |
 | 257 |  12 |   MOSI1 |      | 1 | 19 || 20 |   |      | 0v      |     |     |
 | 256 |  13 |   MISO1 |      | 1 | 21 || 22 | 0 | IN   | GPIO5C3 | 6   | 171 |
 | 254 |  14 |   SCLK1 |      | 1 | 23 || 24 | 1 |      | CE0     | 10  | 255 |
 |     |     |      0v |      |   | 25 || 26 | 1 | OUT  | CE1     | 11  | 251 |
 | 233 |  30 |   SDA.2 |      | 1 | 27 || 28 | 1 |      | SCL.2   | 31  | 234 |
 | 165 |  21 | GPIO5B5 |   IN | 1 | 29 || 30 |   |      | 0v      |     |     |
 | 168 |  22 | GPIO5C0 |   IN | 1 | 31 || 32 | 1 | SERL | GPIO7C7 | 26  | 239 |
 | 238 |  23 | GPIO7C6 | SERL | 1 | 33 || 34 |   |      | 0v      |     |     |
 | 185 |  24 | GPIO6A1 |      | 0 | 35 || 36 | 1 | SERL | GPIO7A7 | 27  | 223 |
 | 224 |  25 | GPIO7B0 | SERL | 1 | 37 || 38 | 0 |      | GPIO6A3 | 28  | 187 |
 |     |     |      0v |      |   | 39 || 40 | 0 |      | GPIO6A4 | 29  | 188 |
 +-----+-----+---------+------+---+----++----+---+------+---------+-----+-----+
 | CPU | wPi |   Name  | Mode | V | Physical | V | Mode | Name    | wPi | CPU |
 +-----+-----+---------+------+---+--Tinker--+---+------+---------+-----+-----+
*/
FILE *fp;
char reads[PATH_MAX];

/* Interrupt21: */
void Interrupt21(void) {
   fp = popen("input keyevent 3", "r");
   pclose(fp);
   printf ("Interrupt 21 / Go Home\n\r");
}

int main (int argc, char **argv) {
    int pinnum = 0;
    int loop = 0;

    if(argc < 3) {
        printf("[%s] [Pin] [HiLoCount]\n", argv[0]);
        printf("example:\t%s 3 10\n", argv[0]);
        printf("        \t%s 0 1\n", argv[0]);
    }
    
    if(argv[1] == NULL) {
        return -1;
    } else {
        pinnum = strtol(argv[1], NULL, 10);
    }
    if(argv[2] == NULL) {
        return -1;
    } else {
        loop = strtol(argv[2], NULL, 10);
    }
    
    printf ("Tinker GPIO Init\n");
    printf ("%d: pinnum=[%d] loop=[%d]\n", __LINE__, pinnum, loop);
    
    wiringPiSetup();
    pinMode (strtol(argv[1], NULL, 10), OUTPUT);

    /*Input GPIO228 / Pin 21/ Header.29*/
    pinMode(21, INPUT);
    pullUpDnControl (21, PUD_UP);

    if ( wiringPiISR (21, INT_EDGE_FALLING, &Interrupt21) < 0 ) {
        printf ("Unable to setup ISR: %s\n", strerror (errno));
        return 1;
    }

    int i = 0;
    for (i =0; i < loop; i++)
    {
        printf ("led on\n");
        digitalWrite (LED, HIGH);
        delay (500);
        printf ("led off\n");
        digitalWrite (LED, LOW);
        delay (500);
     }
    
    return 0;
}
